package org.apache.xindice.core.indexer;

/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * $Id: ValueIndexer.java,v 1.1 2001/12/06 21:00:12 bradford Exp $
 */

import org.apache.xindice.core.*;
import org.apache.xindice.core.Collection;
import org.apache.xindice.core.data.*;
import org.apache.xindice.core.filer.*;
import org.apache.xindice.core.query.*;
import org.apache.xindice.util.*;
import org.apache.xindice.xml.*;
import org.apache.xindice.xml.dom.*;

import java.io.*;
import java.util.*;

import org.w3c.dom.*;

/**
 * ValueIndexer is a basic implementation of the Indexer interface.
 * It is used for maintaining element and element@attribute value
 * indexes.
 */

public final class ValueIndexer extends BTree implements Indexer {
   private static final IndexMatch[] EmptyMatches = new IndexMatch[0];
   private static final Value EmptyValue = new Value(new byte[0]);

   private static final byte MATCHES = 20;
   private static final long MATCH_INFO = -1000;

   private static final String NAME = "name";
   private static final String PATTERN = "pattern";
   private static final String TYPE = "type";
   private static final String PAGESIZE = "pagesize";
   private static final String MAXKEYSIZE = "maxkeysize";

   private static final int STRING = 0;
   private static final int TRIMMED = 1;
   private static final int INTEGER = 2;
   private static final int FLOAT = 3;
   private static final int BYTE = 4;
   private static final int CHAR = 5;
   private static final int BOOLEAN = 6;

   private static final int[] sizes = { -1, -1, 8, 8, 1, 2, 1 };

   private static final String STRING_VAL = "string";
   private static final String TRIMMED_VAL = "trimmed";
   private static final String SHORT_VAL = "short";
   private static final String INT_VAL = "int";
   private static final String LONG_VAL = "long";
   private static final String FLOAT_VAL = "float";
   private static final String DOUBLE_VAL = "double";
   private static final String BYTE_VAL = "byte";
   private static final String CHAR_VAL = "char";
   private static final String BOOLEAN_VAL = "boolean";

   private Configuration config;
   private Collection collection;
   private SymbolTable symbols;

   private String name;
   private String pattern;
   private int type;
   private int typeSize = 32;
   private boolean wildcard = false;

   private FileHeader fileHeader;

   public ValueIndexer() {
      super();
      fileHeader = getFileHeader();
      fileHeader.setPageSize(1024);
   }

   public void setConfig(Configuration config) {
      this.config = config;
      try {
         name = config.getAttribute(NAME);

         pattern = config.getAttribute(PATTERN);
         wildcard = pattern.indexOf('*') != -1;

         // Determine the Index Type
         String tv = config.getAttribute(TYPE, STRING_VAL).toLowerCase();
         if ( tv.equals(STRING_VAL) )       type = STRING;
         else if ( tv.equals(TRIMMED_VAL) ) type = TRIMMED;
         else if ( tv.equals(SHORT_VAL) )   type = INTEGER;
         else if ( tv.equals(INT_VAL) )     type = INTEGER;
         else if ( tv.equals(LONG_VAL) )    type = INTEGER;
         else if ( tv.equals(FLOAT_VAL) )   type = FLOAT;
         else if ( tv.equals(DOUBLE_VAL) )  type = FLOAT;
         else if ( tv.equals(BYTE_VAL) )    type = BYTE;
         else if ( tv.equals(CHAR_VAL) )    type = CHAR;
         else if ( tv.equals(BOOLEAN_VAL) ) type = BOOLEAN;
         else {
            if ( pattern.indexOf('@') != -1 )
               type = STRING;
            else
               type = TRIMMED;
         }

         typeSize = sizes[type];

         fileHeader.setPageSize(config.getIntAttribute(PAGESIZE, fileHeader.getPageSize()));
         fileHeader.setMaxKeySize(config.getShortAttribute(MAXKEYSIZE, fileHeader.getMaxKeySize()));

         setLocation(name);
      }
      catch ( Exception e ) {
         org.apache.xindice.Debug.printStackTrace(e);
      }
   }

   public Configuration getConfig() {
      return config;
   }

   public String getName() {
      return name;
   }

   public void setLocation(String location) {
      setFile(new File(collection.getCollectionRoot(), location+".idx"));
   }

   public void setCollection(Collection collection) {
      try {
         this.collection = collection;
         symbols = collection.getSymbols();
      }
      catch ( Exception e ) {
         org.apache.xindice.Debug.printStackTrace(e);
      }
   }

   public String getIndexStyle() {
      return STYLE_NODEVALUE;
   }

   public String getPattern() {
      return pattern;
   }

   public Value getTypedValue(String value) {
      if ( type != STRING && type != TRIMMED ) {
         value = value.trim();

         if ( value.length() == 0 )
            return EmptyValue;

         byte[] b = new byte[typeSize];
         try {
            switch ( type ) {
               case INTEGER:
                  long l = Long.parseLong(value);
                  b[0] = (byte)((l >>> 56) & 0xFF);
                  b[1] = (byte)((l >>> 48) & 0xFF);
                  b[2] = (byte)((l >>> 40) & 0xFF);
                  b[3] = (byte)((l >>> 32) & 0xFF);
                  b[4] = (byte)((l >>> 24) & 0xFF);
                  b[5] = (byte)((l >>> 16) & 0xFF);
                  b[6] = (byte)((l >>> 8) & 0xFF);
                  b[7] = (byte)((l >>> 0) & 0xFF);
                  break;
               case FLOAT:
                  double d = Double.parseDouble(value);
                  int i1 = (int)Math.round(d);
                  int i2 = (int)Math.round((d-i1)*1000000000);
                  b[0] = (byte)((i1 >>> 24) & 0xFF);
                  b[1] = (byte)((i1 >>> 16) & 0xFF);
                  b[2] = (byte)((i1 >>> 8) & 0xFF);
                  b[3] = (byte)((i1 >>> 0) & 0xFF);
                  b[4] = (byte)((i2 >>> 24) & 0xFF);
                  b[5] = (byte)((i2 >>> 16) & 0xFF);
                  b[6] = (byte)((i2 >>> 8) & 0xFF);
                  b[7] = (byte)((i2 >>> 0) & 0xFF);
                  break;
               case BYTE:
                  b[0] = Byte.parseByte(value);
                  break;
               case CHAR:
                  char c = value.charAt(0);
                  b[0] = (byte)((c >>> 8) & 0xFF);
                  b[1] = (byte)((c >>> 0) & 0xFF);
                  break;
               case BOOLEAN:
                  if ( "[true][yes][1][y][on]".indexOf("["+value.toString().toLowerCase()+"]") != -1 )
                     b[0] = 1;
                  else if ( "[false][no][0][n][off]".indexOf("["+value.toString().toLowerCase()+"]") != -1 )
                     b[0] = 0;
                  else
                     return EmptyValue;
                  break;
            }
            return new Value(b);
         }
         catch ( Exception e ) {
            return EmptyValue;
         }
      }

      if ( type == TRIMMED )
         value = QueryEngine.normalizeString(value);

      return new Value(value);
   }

   private Value getCombinedValue(Key key, int pos, int len, short elemID, short attrID) {
      Value result;
      try {
         int l = key.getLength();
         byte[] b = new byte[l+13];
         
         // Write the key
         System.arraycopy(key.getData(), 0, b, 0, l);
         b[l] = 0;
         
         // Write the pos
         b[l+1] = (byte)((pos >>> 24) & 0xFF);
         b[l+2] = (byte)((pos >>> 16) & 0xFF);
         b[l+3] = (byte)((pos >>> 8)  & 0xFF);
         b[l+4] = (byte)((pos >>> 0)  & 0xFF);

         // Write the len
         b[l+5] = (byte)((len >>> 24) & 0xFF);
         b[l+6] = (byte)((len >>> 16) & 0xFF);
         b[l+7] = (byte)((len >>> 8)  & 0xFF);
         b[l+8] = (byte)((len >>> 0)  & 0xFF);

         // Write the elemID
         b[l+9] = (byte)((elemID >>> 8)  & 0xFF);
         b[l+10] = (byte)((elemID >>> 0) & 0xFF);
         
         // Write the attrID
         b[l+11] = (byte)((attrID >>> 8) & 0xFF);
         b[l+12] = (byte)((attrID >>> 0) & 0xFF);

         result = new Value(b);
      }
      catch ( Exception e ) {
         result = null; // This will never happen
      }
      return result;
   }
   
   private IndexMatch getIndexMatch(Value v) {
      byte[] b = v.getData();
      int l = b.length-13;
      Key key = new Key(b, 0, b.length-13);
      
      int pos = ((b[l+1] << 24) | (b[l+2] << 16) | (b[l+3] << 8) | b[l+4]);
      int len = ((b[l+5] << 24) | (b[l+6] << 16) | (b[l+7] << 8) | b[l+8]);
      short elemID = (short)((b[l+9] << 8) | b[l+10]);
      short attrID = (short)((b[l+11] << 8) | b[l+12]);

      return new IndexMatch(key, pos, len, elemID, attrID);
   }
   
   public synchronized void remove(String value, Key key, int pos, int len, short elemID, short attrID) throws DBException {
      Value v = getTypedValue(value);
      if ( type != STRING && type != TRIMMED && v.getLength() == 0 )
         return;

      try {
         BTreeRootInfo root = findBTreeRoot(v);
         Value cv = getCombinedValue(key, pos, len, elemID, attrID);
         removeValue(root, cv);
      }
      catch ( DBException d ) {
         throw d;
      }
      catch ( Exception e ) {
         org.apache.xindice.Debug.printStackTrace(e);
      }
   }

   public synchronized void add(String value, Key key, int pos, int len, short elemID, short attrID) throws DBException {
      Value v = getTypedValue(value);
      if ( type != STRING && type != TRIMMED && v.getLength() == 0 )
         return;
  
      try {
         BTreeRootInfo root;
         
         try {
            root = findBTreeRoot(v);
         }
         catch ( BTreeNotFoundException e ) {
            root = createBTreeRoot(v);
         }

         Value cv = getCombinedValue(key, pos, len, elemID, attrID);
         addValue(root, cv, MATCH_INFO);
      }
      catch ( DBException d ) {
         throw d;
      }
      catch ( IOException e ) {
         throw new BTreeCorruptException("Corruption detected on add");
      }
      catch ( Exception e ) {
         org.apache.xindice.Debug.printStackTrace(e);
      }
   }

   public synchronized void flush() throws DBException {
      super.flush();
   }
   
   public synchronized IndexMatch[] queryMatches(final IndexQuery query) throws DBException {
      // Pre-process the value-set for typing and trimming
      if ( type != STRING ) {
         Value[] vals = query.getValues();
         for ( int i = 0; i < vals.length; i++ )
            vals[i] = getTypedValue(vals[i].toString());
      }

      // Now issue the query
      final List results = new ArrayList();

      try {
         query(query, new BTreeCallback() {
            public boolean indexInfo(Value value, long pos) {
               try {
                  if ( pos == MATCH_INFO ) {
                     IndexMatch match = getIndexMatch(value);
                     if ( !wildcard )
                        results.add(match);
                     else {
                        IndexPattern pt = new IndexPattern(symbols, match.getElement(), match.getAttribute());
                        if ( pt.getMatchLevel(query.getPattern()) > 0 )
                           results.add(match);
                     }
                  }
                  else {
                     BTreeRootInfo root = new BTreeRootInfo(value, pos);
                     query(root, null, this);
                  }
                  return true;
               }
               catch ( Exception e ) {
                  org.apache.xindice.Debug.printStackTrace(e);
               }
               return true;
            }
         });
      }
      catch ( IOException e ) {
         throw new BTreeCorruptException("Corruption detected on query");
      }
      catch ( Exception e ) {
         org.apache.xindice.Debug.printStackTrace(e);
      }

      return (IndexMatch[])results.toArray(EmptyMatches);
   }
}


