package org.apache.xindice.server;

/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * $Id: Kernel.java,v 1.2 2002/02/26 07:10:09 kstaken Exp $
 */
 
import org.apache.xindice.util.*;

import org.w3c.dom.*;
import org.xml.sax.*;
import javax.xml.parsers.*;

import java.io.*;
import java.util.*;

/**
 * The Kernel class is the glue that holds the Xindice information
 * system together. It is the central processing unit of the entire
 * Xindice system. Kernel does almost no processing of its own. The
 * kernel's main job is to load in the various API managers and to start
 * the system services.
 */

public final class Kernel implements Runnable {
   private static final String NAME = "name";
   private static final String CLASS = "class";
   private static final String COMMON = "common";
   private static final String SERVER = "server";
   private static final String KEEPALIVE = "keepalive";
   private static final String GCINTERVAL = "gcinterval";
   private static final String SERVICES = "services";
   private static final String FILES = "files";
   private static final String LOGGING = "logging";
   private static final String RESOURCES = "resources";
   private static final String USERS = "users";
   private static final String SCRIPTS = "scripts";
   private static final String COMPONENTS = "components";

   private Configuration config = null;
   private Configuration commonConfig = null;
   private List tasks = new ArrayList(); // Of TaskInfo
   private Task finalizer = null;
   private Thread scheduler = null;
   private boolean running = false;
   private String serverName = null;

   /* Kernel API manager hooks */

   private FileManager files;
   private ResourceManager resources;
   private ScriptManager scripts;
   private LogManager logs;
   private ServiceManager services;
   private ComponentManager components;
   private UserManager users;

   public Kernel() {
      this("system.xml", true);
   }

   public Kernel(final String name) {
       this(name, true);
   }
   
   public Kernel(final String name, boolean exit) {

       //System.runFinalizersOnExit(true);

      // Read in the configuration
      Document tmp = null;
      try {
         DocumentBuilder db = DocumentBuilderFactory.newInstance().newDocumentBuilder();
         FileInputStream fis = new FileInputStream(name);
         BufferedInputStream bis = new BufferedInputStream(fis, 4096);
         tmp = db.parse(fis);
         fis.close();
      }
      catch ( Exception e ) {
         System.err.println("\u0007FATAL ERROR: Reading configuration file '"+name+"'");
         org.apache.xindice.Debug.printStackTrace(e);
         if ( exit ) {
             System.exit(1);    
         }
         else {
             throw new RuntimeException("\u0007FATAL ERROR: Reading configuration file '"+name+"'");
         }
      }
      final Document doc = tmp;
      config = new Configuration(doc.getDocumentElement(), false); // Read/Write

      commonConfig = config.getChild(COMMON);
      Configuration serverConfig = config.getChild(SERVER);

      // Load And Initialize The Kernel API Modules
      services = (ServiceManager)loadAPI(serverConfig.getChild(SERVICES));
      files = (FileManager)loadAPI(serverConfig.getChild(FILES));
      logs = (LogManager)loadAPI(serverConfig.getChild(LOGGING));
      resources = (ResourceManager)loadAPI(serverConfig.getChild(RESOURCES));
      users = (UserManager)loadAPI(serverConfig.getChild(USERS));
      scripts = (ScriptManager)loadAPI(serverConfig.getChild(SCRIPTS));
      components = (ComponentManager)loadAPI(serverConfig.getChild(COMPONENTS));

      if ( !startServices() ) {          
         System.err.println("\u0007FATAL ERROR: Service manager could not be started");
          if ( exit ) {
             System.exit(1);    
          }
          else {
             throw new RuntimeException("\u0007FATAL ERROR: Service manager could not be started");
          }
      }

      // Start The Scheduling Daemon
      scheduler = new Thread(this, "Task Scheduler");
      scheduler.setDaemon(!serverConfig.getBooleanAttribute(KEEPALIVE));
      scheduler.start();
      
      // Garbage Collect And Finalize Every Specified Number Of Seconds
      long gcinterval = serverConfig.getIntAttribute(GCINTERVAL)*1000;
      if ( gcinterval > 0 ) {
         finalizer = new Task() {
            public void runTask() {
               System.gc();
               System.runFinalization();
            }
         };
         addTask(finalizer, gcinterval);
      }

      System.out.println();
      System.out.println("Server Running");
      running = true;
   }

   /**
    * run is used to keep the kernel alive with no services running
    * and to run interval-scheduled tasks
    */
   public void run() {
      TaskInfo info;
      long wait = 5000;
      long time = 0;
      long done = 0;
      long diff = 0;
      int size = 0;
      int i = 0;
      while ( true ) {
         try {
            wait = 5000;
            size = tasks.size();
            for ( i = 0; i < size; i++ ) {
               info = (TaskInfo)tasks.get(i);
               time = System.currentTimeMillis();
               if ( time >= info.lastrun+info.interval ) {
                  try {
                     info.task.runTask();
                  }
                  catch ( Exception e ) {
                     if ( info.error == false ) {
                        info.error = true;
                        System.err.println("\u0007TASK ERROR: "+e);
                     }
                  }
                  done = System.currentTimeMillis();
                  info.lastrun = time;
                  diff = done-time;
                  if ( info.interval-diff < wait )
                     wait = info.interval-diff;
               }
               else {
                  diff = (info.lastrun+info.interval)-time;
                  if ( diff > 0 && diff < wait )
                     wait = diff;
               }
               //Thread.yield();
            }
            Thread.sleep(wait);
         }
         catch ( Exception e ) {
         }
      }
   }

   public synchronized void addTask(Task task, long interval) {
      int size = tasks.size();
      for ( int i = 0; i < size; i++ ) {
         TaskInfo info = (TaskInfo)tasks.get(i);
         if ( task == info.task ) {
            info.interval = interval;
            return;
         }
      }
      tasks.add(new TaskInfo(task, interval));
   }

   public synchronized void removeTask(Task task) {
      int size = tasks.size();
      for ( int i = 0; i < size; i++ ) {
         TaskInfo info = (TaskInfo)tasks.get(i);
         if ( task == info.task ) {
            tasks.remove(i);
            return;
         }
      }
   }

   public static void disposeOf(Object obj) {
      if ( obj != null && (obj instanceof Disposable) )
         ((Disposable)obj).dispose();
   }

   public void setKernelOf(Object obj) {
      if ( obj != null && (obj instanceof KernelAccess) )
         ((KernelAccess)obj).setKernel(this);
   }

   public Configuration getCommonConfig() {
      return commonConfig;
   }

   public void shutDown(int exitCode) {
       shutDown(exitCode, true);
   }

   /*
    * Shutdown the kernel. If exit is false the VM won't exit but the current kernel
    * instance becomes invalid and should no longer be used.
    */
   public void shutDown(int exitCode, boolean exit) {

      System.out.println();
      
      removeTask(finalizer);
      services.stopServices();

      disposeOf(components);
      disposeOf(scripts);
      disposeOf(users);
      disposeOf(resources);
      disposeOf(logs);
      disposeOf(files);
      disposeOf(services);

      components = null;
      scripts = null;
      users = null;
      resources = null;
      logs = null;
      files = null;
      services = null;

      System.out.println("Server shutdown (Exit code "+exitCode+")");

      // Collect Garbage And Perform Disposal
      System.gc();
      System.runFinalization();

      // Shut Down The Kernel
      if ( exit ) {
          System.exit(exitCode);    
      }      
   }

   private Object loadAPI(Configuration config) {
      if (config == null) {
         return null;
      }
      
      Object obj = null;
      String className = config.getAttribute(CLASS);
      if (className == null) {
         String name = config.getAttribute(NAME);
         System.err.println("\u0007FATAL ERROR: Reading API module '" + name +
            "' no class attribute specified");
      }
      
      try {
         obj = Class.forName(className).newInstance();
         setKernelOf(obj);
         if ( obj instanceof Configurable )
            ((Configurable)obj).setConfig(config);
      }
      catch ( Exception e ) {
         System.err.println("\u0007FATAL ERROR: Reading API module '"+className+"'");
         System.exit(1);
      }
      return obj;
   }

   public boolean isRunning() {
      return running;
   }

   /* Stub methods to the API managers */

   /* FileManager stubs */

   public FileResource getFileResource(String filename) {
      return files.getFileResource(filename);
   }

   public FileResource getTemporaryFileResource() {
      return files.getTemporaryFileResource();
   }

   /* ResourceManager stubs */

   public void setResource(Versioning versioning, String name, Object object) throws ParameterException {
      resources.setResource(versioning, name, object);
   }

   public void setResource(String name, Object object) throws ParameterException {
      resources.setResource(name, object);
   }

   public void removeResource(Versioning versioning, String name) {
      resources.removeResource(versioning, name);
   }

   public void removeResource(String name) {
      resources.removeResource(name);
   }

   public Object getResource(Versioning versioning, String name) {
      return resources.getResource(versioning, name);
   }

   public Object getResource(String name) {
      return resources.getResource(name);
   }

   public void lockResource(Versioning versioning, String name) {
      resources.lockResource(versioning, name);
   }

   public void lockResource(String name) {
      resources.lockResource(name);
   }

   public void unlockResource(Versioning versioning, String name) {
      resources.unlockResource(versioning, name);
   }

   public void unlockResource(String name) {
      resources.unlockResource(name);
   }

   /* ScriptManager stubs */

   public Script getScript(String name) {
      return scripts.getScript(name);
   }

   public void removeScript(String name) {
      scripts.removeScript(name);
   }

   public void runScript(String name, Gateway gateway, boolean useFilters) {
      scripts.runScript(name, gateway, useFilters);
   }

   public void runScript(Script script, Gateway gateway, boolean useFilters) {
      scripts.runScript(script, gateway, useFilters);
   }

   public boolean addScript(String name, Script script) {
      return scripts.addScript(name, script);
   }

   /* LogManager Stubs */

   public void logMessage(int type, String host, String user, String message) {
      logs.logMessage(type, host, user, message);
   }
   
   public void logMessage(int type, String host, String message) {
      logs.logMessage(type, host, message);
   }
   
   public void logMessage(int type, String message) {
      logs.logMessage(type, message);
   }

   /* ComponentManager stubs */

   public boolean addComponent(ScriptComponent component, String name) {
      return components.addComponent(component, name);
   }

   public boolean addComponent(ScriptComponent component) {
      return components.addComponent(component);
   }

   public ScriptComponent getComponent(String name) {
      return components.getComponent(name);
   }

   public ScriptFilter[] getFilters(Versioning versioning) {
      return components.getFilters(versioning);
   }

   public boolean removeComponent(String name) {
      return components.removeComponent(name);
   }

   /* ServiceManager stubs */

   public boolean startServices() {
      return services.startServices();
   }

   public void stopServices() {
      services.stopServices();
   }

   public boolean addService(Service service) {
      return services.addService(service);
   }

   public boolean removeService(Service service) {
      return services.removeService(service);
   }

   public String[] listServices() {
      return services.listServices();
   }

   public Service getService(String name) {
      return services.getService(name);
   }

   /* UserManager stubs */

   public User getUser(String username) {
      return users.getUser(username);
   }

   public User newUser() {
      return users.newUser();
   }

   public void setUser(User user) {
      users.setUser(user);
   }

   public void deleteUser(String username) {
      users.deleteUser(username);
   }

   public String[] listUsers(String startsWith) {
      return users.listUsers(startsWith);
   }

   public String[] listUsers() {
      return users.listUsers();
   }

   /**
    * TaskInfo
    */

   private class TaskInfo {
      public Task task;
      public long interval = 0;
      public long lastrun = 0;
      public boolean error = false;

      public TaskInfo(Task task, long interval) {
         this.task = task;
         this.interval = interval;
         this.lastrun = System.currentTimeMillis();
      }
   }
   
   
   /**
    * TextWriter takes a Document, DocumentFragment, or Element and streams it
    * as text to an output source (or a String) -- Stolen from org.apache.xindice.xml
    */
   
   public final class TextWriter {
      private Node node = null;
   
      public TextWriter(Node node) throws DOMException {
         this.node = node;
      }
   
      private void writeNode(Writer writer, Node node) throws IOException {
         short type = node.getNodeType();
         switch ( type ) {
   
            case Node.DOCUMENT_NODE: {
               writer.write("<?xml version=\"1.0\"?>\n");
               writeChildren(writer, node);
               break;
            }
   
            case Node.DOCUMENT_FRAGMENT_NODE: {
               writeChildren(writer, node);
               break;
            }

            case Node.ELEMENT_NODE: {
               Element e = (Element)node;
               String n = e.getTagName();
   
               writer.write('<');
               writer.write(n);
   
               NamedNodeMap a = e.getAttributes();
               int size = a.getLength();
               for ( int i = 0; i < size; i++ ) {
                  Attr att = (Attr)a.item(i);
                  writer.write(' ');
                  writeNode(writer, att);
               }
   
               if ( e.hasChildNodes() ) {
                  writer.write('>');
                  writeChildren(writer, node);
                  writer.write("</");
                  writer.write(n);
                  writer.write('>');
               }
               else
                  writer.write(" />");
               break;
            }
   
            case Node.ATTRIBUTE_NODE:
               Attr a = (Attr)node;
               writer.write(a.getName());
               writer.write("=\"");
               writeEscapedText(writer, a.getValue());
               writer.write("\"");
               break;
   
   
            case Node.PROCESSING_INSTRUCTION_NODE: {
               ProcessingInstruction pi = (ProcessingInstruction)node;
               writer.write("<?");
               writer.write(pi.getTarget());
               writer.write(" ");
               writer.write(pi.getData());
               writer.write("?>\n");
               break;
            }
   
            case Node.TEXT_NODE: {
               writeEscapedText(writer, node.getNodeValue());
               break;
            }
   
            case Node.CDATA_SECTION_NODE: {
               writer.write("<![CDATA[");
               writer.write(node.getNodeValue());
               writer.write("]]>");
               break;
            }
   
            case Node.COMMENT_NODE: {
               writer.write("<!--");
               writer.write(node.getNodeValue());
               writer.write("-->");
               break;
            }
         }
      }
   
      private void writeChildren(Writer writer, Node node) throws IOException {
         NodeList l = node.getChildNodes();
         int size = l.getLength();
         for ( int i = 0; i < size; i++ )
            writeNode(writer, l.item(i));
      }
   
      private void writeEscapedText(Writer writer, String text) throws IOException {
         char[] value = text.toCharArray();
         String outval = null;
         int start = 0;
         int len = 0;
         for ( int i = 0; i < value.length; i++ ) {
            switch ( value[i] ) {
               case '&'  : outval = "&amp;";  break;
               case '\'' : outval = "&apos;"; break;
               case '\"' : outval = "&quot;"; break;
               case '<'  : outval = "&lt;";   break;
               case '>'  : outval = "&gt;";   break;
               default   : len++;             break;
            }
   
            if ( outval != null ) {
               if ( len > 0 )
                  writer.write(value, start, len);
               writer.write(outval);
               start = i+1;
               len = 0;
               outval = null;
            }
         }
         if ( len > 0 )
            writer.write(value, start, len);
      }
      
      /**
       * write writes the node to the OutputStream as text.
       *
       * @param output The OutputStream to write to
       */
      public void write(OutputStream output) throws IOException {
         try {
            OutputStreamWriter o = new OutputStreamWriter(output);
            BufferedWriter buf = new BufferedWriter(o, 4096);
            writeNode(buf, node);
            buf.flush();
         }
         catch ( Exception e ) {
             org.apache.xindice.Debug.printStackTrace(e);
         }
      }
   }
}

