package org.apache.xindice.server.methods;

/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * $Id: GET.java,v 1.1.1.1 2001/12/06 19:33:56 bradford Exp $
 */

import org.apache.xindice.server.*;
import org.apache.xindice.util.*;
import java.io.*;
import java.util.*;

/**
 * GET is an HTTP 1.1 compliant GET method. This Script performs
 * native socket output for maximum performance and checks the
 * If-Modified-Since header to determine whether to send the content or not.
 */

public final class GET extends SimpleScript {
   private static final String INDEX = "index";
   private static final String BROWSE = "browse";
   private static final String DEFAULTICON = "defaulticon";
   private static final String FOLDERICON = "foldericon";
   private static final String FOLDERUPICON = "folderupicon";
   private static final String SPACERICON = "spacericon";
   private static final String MIME = "mime";
   private static final String MIMETYPE = "mimetype";
   private static final String EXT = "ext";
   private static final String TYPE = "type";
   private static final String DEFAULT = "default";
   private static final String ICON = "icon";
   private static final String LOCATION = "location";

   private String docIndex = null;
   private boolean autoDir = false;
   private String fileImage = null;
   private String folderImage = null;
   private String folderUpImage = null;
   private String spacerImage = null;
   private Map iconTypes = null; // String
   private Map mimeTypes = null; // String
   private String defaultMime = null;

   public void setConfig(Configuration config) throws XindiceException {
      super.setConfig(config);
      docIndex = config.getAttribute(INDEX);
      autoDir = config.getBooleanAttribute(BROWSE);
      fileImage = config.getAttribute(DEFAULTICON);
      folderImage = config.getAttribute(FOLDERICON);
      folderUpImage = config.getAttribute(FOLDERUPICON);
      spacerImage = config.getAttribute(SPACERICON);

      mimeTypes = new HashMap();
      Configuration mimeConfig = kernel.getCommonConfig().getChild(MIME);

      mimeConfig.processChildren(MIMETYPE,
         new ConfigurationCallback() {
            public void process(Configuration cfg) {
               StringTokenizer st = new StringTokenizer(cfg.getAttribute(EXT));
               while ( st.hasMoreTokens() )
                  mimeTypes.put(st.nextToken(), cfg.getAttribute(TYPE));
            }
      });

      defaultMime = mimeConfig.getAttribute(DEFAULT);

      iconTypes = new HashMap();
      config.processChildren(ICON,
         new ConfigurationCallback() {
            public void process(Configuration cfg) {
               iconTypes.put(cfg.getAttribute(EXT), cfg.getAttribute(LOCATION));
         }
      });
   }

   private void sendNotFoundMessage(Gateway gateway, String pathinfo) {
      String message = "File Not Found: "+pathinfo;
      gateway.logMessage(LogManager.LOG_ERROR, message);
      gateway.sendError(404, message);
   }

   private String formatField(String field, String endTag, int length) {
      if ( field.length()>length )
         field = field.substring(0, length);
      StringBuffer result = new StringBuffer(StringUtilities.leftJustify(field, length));
      if ( endTag != null )
         result.insert(field.length(), endTag);
      return result.toString();
   }

   private void sendDirectoryStructure(Gateway gateway, String path) {
      FileResource file;
      int idx = 0;
      int i = 0;
      String url;
      String pathinfo = gateway.getRequestHeader("PATH_INFO");
      path = path.substring(0, path.lastIndexOf("/"+docIndex));
      file = gateway.getFileResource(path);

      PrintStream out = new PrintStream(gateway.getOutputStream());
      gateway.setResponseHeader("Content-type", "text/html");

      out.println("<HTML>\n<HEAD>\n<TITLE>Contents of folder: "+pathinfo+"</TITLE>\n</HEAD>\n<BODY BGCOLOR=\"#FFFFFF\">");
      out.println("<FONT FACE=\"Helvetica,Helv,Arial\"><STRONG>Contents Of Folder: "+pathinfo+"</STRONG></FONT>");
      out.println("<PRE><IMG SRC=\""+folderUpImage+"\" BORDER=\"0\" ALIGN=\"BASELINE\"/> <A HREF=\"..\">Parent Folder</A>");
      FileResource[] names = file.getChildren();
      if ( names == null ) {
         sendNotFoundMessage(gateway, path);
         return;
      }
      int len = names.length;
      for ( i = 0; i < len; i++ ) {
         file = names[i];
         out.print("<IMG SRC=\""+spacerImage+"\">");
         if ( file.isDirectory() ) {
            out.print("<IMG SRC=\""+folderImage+"\" BORDER=\"0\" ALIGN=\"BASELINE\"/>");
            out.print(" <A HREF=\""+file.getName()+"/\">");
            out.print(formatField(file.getName(), "</A>", 45));
            out.print(" <I>"+formatField("Folder", "</I>", 10));
         }
         else {
            idx = file.getName().lastIndexOf('.');
            url = null;
            if ( idx != -1 )
               url = (String)iconTypes.get(file.getName().substring(idx).toLowerCase());
            if ( url == null )
               url = fileImage;
            out.print("<IMG SRC=\""+url+"\" BORDER=\"0\" ALIGN=\"BASELINE\"/>");
            out.print(" <A HREF=\""+file.getName()+"\">");
            out.print(formatField(file.getName(),"</A>", 45));
            out.print(" "+formatField(""+file.length(), null, 10));
         }
         out.print(" "+DateUtilities.getGMTString(new Date(file.lastModified())));
         out.println();
      }
      out.println("</PRE>\n</BODY>\n</HTML>");
      gateway.send(200);
   }

   private boolean isNotModified(Gateway gateway, Date modified, String since) {
      if ( modified != null && since.length() > 0 ) {
         StringTokenizer st = new StringTokenizer(since, ";");
         if ( st.hasMoreTokens() )
            since = st.nextToken().trim();
         try {
            Date comp = DateUtilities.getGMTDate(since);
            if ( modified.equals(comp) ) {
               gateway.resetOutputStream();
               gateway.send(304);
               return true;
            }
         }
         catch ( Exception e ) {
         }
      }
      return false;
   }

   private void setContentType(Gateway gateway, String path) {
      int idx = path.lastIndexOf('.');
      if ( idx != -1 ) {
         String type = (String)mimeTypes.get(path.substring(idx+1).toLowerCase());
         if ( type == null )
            type = defaultMime;
         gateway.setResponseHeader("Content-type", type);
      }
   }

   public void run(Gateway gateway) {
      int code = 200;
      String method = gateway.getRequestHeader("REQUEST_METHOD");
      String since = gateway.getRequestHeader("If-Modified-Since");
      String pathinfo = gateway.getRequestHeader("PATH_INFO");
      get: {
         String path = gateway.getScriptPath();
         if ( pathinfo.endsWith("/") )
            path += docIndex;
         else if ( path.endsWith("/") ) {
            gateway.redirectLocal(pathinfo+'/');
            return;
         }
         try {
            FileResource check = gateway.getFileResource(path);
            if ( check.isDirectory() ) {
               gateway.redirectLocal(pathinfo+'/');
               return;
            }
            else if ( check.exists() ) {
               Date modified = new Date(check.lastModified());
               if ( isNotModified(gateway, modified, since) ) {
                  code = 304;
                  break get;
               }
               setContentType(gateway, path);
               gateway.removeResponseHeader("Cache-control");
               gateway.setResponseHeader("Content-length", ""+check.length());
               gateway.setResponseHeader("Last-modified", DateUtilities.getGMTString(modified));

               OutputStream os = gateway.getNativeOutputStream();
               gateway.getHeaderContent(200).writeTo(os);
               if ( !method.equals("HEAD") )
                  check.copyStreamTo(os);
            }
            else {
               if ( path.endsWith('/'+docIndex) && autoDir )
                  sendDirectoryStructure(gateway, path);
               else {
                  sendNotFoundMessage(gateway, path);
                  return;
               }
            }
         }
         catch ( Exception e ) {
            sendNotFoundMessage(gateway, path);
            return;
         }
      }
      gateway.logMessage(LogManager.LOG_GENERAL, method+" "+gateway.getScriptPath()+" "+code+" "+gateway.getResponseHeader("Content-length"));
   }
}


