package org.apache.xindice.server.standard;

/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * $Id: StdScriptManager.java,v 1.1.1.1 2001/12/06 19:33:56 bradford Exp $
 */

import org.apache.xindice.server.*;
import org.apache.xindice.util.*;
import java.io.*;
import java.util.*;

/**
 * StdScriptManager
 */

public final class StdScriptManager implements Configurable, ScriptManager, KernelAccess, Disposable {
   private static final String SCRIPTDIR = "scriptdir";
   private static final String EXCEPTIONS = "exceptions";
   private static final String STABLESCRIPTS = "stablescripts";
   
   private Kernel kernel;
   private Configuration config;
   private Map scripts = Collections.synchronizedMap(new HashMap()); // of Script
   private DirectoryClassLoader loader;
   private String scriptDir;
   private boolean displayExceptions = false;
   private boolean stableScripts = false;

   public void setKernel(Kernel kernel) {
      this.kernel = kernel;
   }
   
   public void setConfig(Configuration config) {
      this.config = config;
      scriptDir = config.getAttribute(SCRIPTDIR);
      displayExceptions = config.getBooleanAttribute(EXCEPTIONS);
      stableScripts = config.getBooleanAttribute(STABLESCRIPTS);
      loader = new DirectoryClassLoader(scriptDir);
   }
   
   public Configuration getConfig() {
      return config;
   }

   public boolean addScript(String name, Script script) {
      Script old = (Script)scripts.get(name);
      if ( old == null ) {
         scripts.put(name, script);
         return true;
      }
      return false;
   }

   public void removeScript(String name) {
      Kernel.disposeOf(scripts.remove(name));
   }

   public Script getScript(String name) {
      Script script = (Script)scripts.get(name);
      if ( stableScripts && script != null )
         return script;
      if ( script != null ) {
         synchronized(this) {
            if ( loader.isModified(name) ) {
               loader = new DirectoryClassLoader(scriptDir);
               kernel.disposeOf(script);
               scripts.remove(name);
               script = null;
            }
         }
      }
      if ( script == null ) {
         try {
            Object obj = loader.loadClass(name, true).newInstance();
            if ( obj instanceof Script ) {
               script = (Script)obj;
               scripts.put(name, script);
            }
         }
         catch ( Exception e ) {
            script = null;
         }
      }
      return script;
   }

   public void runScript(String name, Gateway gateway, boolean useFilters) {
      Script script = getScript(name);
      if ( script != null )
         runScript(script, gateway, useFilters);
      else {
         gateway.logMessage(LogManager.LOG_EXCEPTION, "Script Not Found: "+name);
         gateway.sendError(404, "Could Not Find Script: "+name);
      }
   }

   public void runScript(Script script, Gateway gateway, boolean useFilters) {
      boolean shorted = false;
      ScriptFilter filter;
      if ( useFilters ) {
         ScriptFilter[] filters = kernel.getFilters(script);
         for ( int i = 0; i < filters.length; i++ ) {
            filter = filters[i];
            try {
               shorted = !filter.run(script, gateway);
            }
            catch ( Exception e ) {
               shorted = true;
               gateway.logMessage(LogManager.LOG_EXCEPTION, "Error Executing Filter "+filter.getName());
               gateway.sendError(500, "Error Executing Filter: "+filter.getName());
            }
            if ( shorted )
               break;
         }
      }
      if ( !shorted ) {
         try {
            script.run(gateway);
         }
         catch ( Exception e ) {
            gateway.logMessage(LogManager.LOG_EXCEPTION, "Error Executing Script "+script.getName());
            if ( displayExceptions )
               displayException(gateway, script, e);
            else
               gateway.sendError(500, "Error Executing Script: "+script.getName());
         }
      }
   }

   private void displayException(Gateway gateway, Script script, Exception e) {
      gateway.sendException(script.getName(), e);
   }

   public void dispose() {
      Iterator enum = scripts.values().iterator();
      while ( enum.hasNext() )
         Kernel.disposeOf(enum.next());
   }
}
