/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 */

package org.apache.xindice.tools.command;

import org.xml.sax.ContentHandler;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import java.io.StringWriter;
import java.util.HashMap;
import java.util.Map;
import java.util.Stack;
import org.xml.sax.ext.LexicalHandler;

/**
 * SAX content handler that produces a string from the SAX events it receives.
 * After calling endDocument(), the string becomes available by calling
 * <code>toString()</code>
 *
 * @author james.bates@amplexor.com
 * @version 1
 */
public class StringSerializer implements ContentHandler, LexicalHandler {

    /* map that contains a stack for each prefix used in XML document */
    private Map namespaceMap;
    
    /* namespace-declarations that should be written when the next element
       opens */
    private StringBuffer namespaceDecls;

    /* Are we currently inside a CDATA section? */
    private boolean inCDATA;
    
    /* the encoding name that should be written into the XML declaration. May be
     * <code>null</code> indicating no encoding should be written.
     */
    private String outputEncoding;
    
    /* the document so far */
    private StringBuffer outputXml;
    
    /**
     * Creates new <code>StringSerializer</code>.
     *
     * @param outputEncoding the encoding that should be written into the
     * XML declaration for the document. This encoding is not used in any other
     * way: when the resultant string is written to a file, the writing application
     * must still take care to actually write using the correct encoding.
     *
     * If <code>outputEncoding</code> is <code>null</code>, the encoding will
     * be omitted from the XML declaration.
     */
    public StringSerializer(String outputEncoding) {

        this.outputEncoding = outputEncoding;
        outputXml = new StringBuffer();
        namespaceMap = new HashMap();
        namespaceDecls = new StringBuffer();
        inCDATA = false;
    }
    
    public void startDocument() throws org.xml.sax.SAXException {
        
        outputXml.append("<?xml version=\"1.0\"");
        if (outputEncoding != null) {
            
            outputXml.append(" encoding=\"");
            outputXml.append(outputEncoding);
            outputXml.append('\"');
        }
        outputXml.append("?>");
    }
    
    public void endDocument() throws org.xml.sax.SAXException {
    }    

    public void processingInstruction(String target, String data) {

        outputXml.append("<?");
        outputXml.append(target);
        outputXml.append(' ');
        outputXml.append(data);
        outputXml.append("?>");
    }
    
    public void comment(char[] data, int start, int len) {
        
        outputXml.append("<!--");
        outputXml.append(data, start, len);
        outputXml.append("-->");
    }
    
    private void xmlIze(StringBuffer sb, char[] data, int start, int len,
            boolean isAttValue) {
            
        for (int i=start; i < start+len; i++) {
            
            if (data[i]=='<') {
                
                sb.append("&lt;");
            } else if (data[i]=='&') {
                
                sb.append("&amp;");
            } else if ((data[i]=='"') && (isAttValue)) {
                
                sb.append("&quot;");
            } else {
                
                sb.append(data[i]);
            }
        }
    }
    
    private void writeAttribute(StringBuffer sb, String attName, String attValue) {

        sb.append(' ');
        sb.append(attName);
        sb.append("=\"");
        xmlIze(sb, attValue.toCharArray(), 0, attValue.length(), true);
        sb.append('\"');
    }
    
    public void startPrefixMapping(String prefix, String uri) {

        Stack uriStack = (Stack) namespaceMap.get(prefix);
        if (uriStack == null) {
            
            uriStack = new Stack();
            namespaceMap.put(prefix, uriStack);
        }
        
        uriStack.push(uri);

        if (!prefix.equals("")) {
            
            writeAttribute(namespaceDecls, "xmlns:" + prefix, uri);
        } else {
            
            writeAttribute(namespaceDecls, "xmlns", uri);
        }
    }
        
    public void endPrefixMapping(String prefix) {

        Stack uriStack = (Stack) namespaceMap.get(prefix);
        uriStack.pop();
    }
    
    
    public void ignorableWhitespace(char[] data, int start, int len) {

        outputXml.append(data, start, len);
    }    

    private String getPrefix(String qName) {
     
        if (qName.indexOf(':') != -1) {
            
            return qName.substring(0, qName.indexOf(':'));
        } else {
            
            return "";
        }
    }
    
    public void startElement(String namespaceUri, String localName, String qName,
            Attributes att) {

        /* First: check element name qualification */
        String prefix = getPrefix(qName);
        
        if ((!(prefix.equals("") || prefix.equals("xml")))
            && (namespaceMap.get(prefix) == null)) {
            
            writeAttribute(namespaceDecls, "xmlns:" + prefix, namespaceUri);
        }
        
        outputXml.append('<');
        outputXml.append(qName);
        outputXml.append(namespaceDecls);
        namespaceDecls = new StringBuffer();
        
        for (int i=0; i < att.getLength(); i++) {
            
            String attNamespaceUri = att.getURI(i);
            String attLocalName = att.getLocalName(i);
            String attQName = att.getQName(i);
            String attValue = att.getValue(i);
            
            String attPrefix = getPrefix(attQName);
            if (!(attPrefix.equals("") || attPrefix.equals("xml"))) {
                
                if (namespaceMap.get(attPrefix) == null) {
                    
                    writeAttribute(outputXml, "xmlns:" + attPrefix, attNamespaceUri);
                }
            }
            
            writeAttribute(outputXml, attQName, attValue);
        }
        
        outputXml.append('>');
    }
    
    public void endElement(String namespaceUri, String localName, String qName) {
        
        outputXml.append("</");
        outputXml.append(qName);
        outputXml.append('>');
    }
    
    public void skippedEntity(String entity) {
        
        if (entity.startsWith("%")) {
            
            outputXml.append(entity);
            outputXml.append(';');
        } else {
            
            outputXml.append('&');
            outputXml.append(entity);
            outputXml.append(';');
        }
    }
    
    public void characters(char[] data, int start, int len) {

        if (inCDATA) {
            
            outputXml.append(data, start, len);
        } else {
            
            xmlIze(outputXml, data, start, len, false);
        }
    }
    
    public void setDocumentLocator(org.xml.sax.Locator locator) {
    }

    public void startCDATA() {
        
        outputXml.append("<![CDATA[");
        inCDATA = true;
    }
    
    public void endCDATA() throws org.xml.sax.SAXException {

        inCDATA = false;
        outputXml.append("]]>");
    }    
    
   
    public void startEntity(String entity) {
    }
     
    public void endEntity(String entity) {
    }
    
    public void startDTD(String docType, String systemID, String publicID) {
    }
    
    public void endDTD() {
    }

    public String toString() {
        
        return new String(outputXml);
    }
}
